; references
;; All objects are meant to be read-only, but for explicit references
;;
;; We trust the program not to modify anything but this emulated type...

(provide 'ref)
(require 'lib)
(require 'tags)
(require 'vectree)
(require 'types)

(define ref$$tag (types:make-tag 'ref))

(define (make-ref obj)
   "make a variable reference object,
that be initialized to point to object obj"
  (vector ref$$tag obj))
(define ref:new (make-ref #f))
(define ref? (vectree-detector ref$$tag 1))

(define ref:get v-1st)
(define ref:set! v-set-1st!)

(define (ref:map! r f)
   "replaces the referenced element by its image by a given function"
   ($assert ref? r)
   ; make it atomic wrt GC
   (ref:set! r (f (ref:get r))))

(define (ref:postmap! r f)
   "replaces the referenced element by its image by a given function"
   "and returns the previous value"
   ($assert ref? r)
   ; make it atomic wrt GC
   (let ((v (ref:get r)))
   (ref:set! r (f v))
   v))

(define (ref:apply! r f . x)
   (ref:map! r (lambda (v) (apply f v x))))

(define (ref:postapply! r f . x)
   (ref:map! r (lambda (v) (apply f v x))))

(define (ref:post-inc r)
   (ref:postmap! r add1))

(define (ref:pre-inc r)
   (ref:map! r add1))

(define (ref:post-dec r)
   (ref:postmap! r sub1))

(define (ref:pre-dec r)
   (ref:map! r sub1))

;;;;; Counters
;(define (new-counter)
;  (let ((cnt (ref:new 0)))
;  (lambda () (ref:post-inc cnt))))

(types:register 'ref ref?)
